from xpudp import XPConnector
from time import sleep

with XPConnector('127.0.0.1', 49000, 49001) as conn:
    conn.send_command('sim/flight_controls/flaps_down')
    sleep(3)
    conn.send_command('sim/flight_controls/flaps_up')
    sleep(2)
    conn.set_dataref('sim/cockpit/autopilot/heading_mag', 180)
    sleep(2)
    conn.set_dataref('sim/cockpit/autopilot/heading_mag', 0)

    #conn.subscribe_to_dataref('sim/cockpit/autopilot/heading_mag')
    while (val := conn.get_dataref('sim/cockpit/autopilot/heading_mag')) < 180:
        print(f'Heading {val} is less than 180, turn the heading knob to a larger value.')
        sleep(1)
    print(f'Done, heading {val}')



import socket
import struct

from itertools import batched
import time

HOST = ('127.0.0.1', 49000)
HOST_RECV = ('127.0.0.1', 49001)

_drefs = []

def send_command(sock, cmd: str):
    message = b'CMND\x00' + cmd.encode()
    res = sock.sendto(message, HOST)
    return res

def set_dataref(sock, dref, value, var_type='f'):
    '''
    var_type
    ---------------+---------
    char           | c
    unsigned char  | b
    bool           | ?
    short          | h
    unsigned short | H
    int            | i
    unsigned int   | I
    long           | l
    unsigned long  | L
    long long      | q
    unsigned ll    | Q
    float          | f
    double         | d
                   |  

    https://docs.python.org/3/library/struct.html#format-characters
    '''
    message = b'DREF\x00' + struct.pack('=' + var_type, value) + dref.encode() + b'\0xx'
    message = message + b' ' * (509 - len(message))
    res = sock.sendto(message, HOST)
    return res

def subscribe_to_dref(sock, dref, freq=1):
    global _drefs

    if dref not in _drefs:
        _drefs.append(dref)
    idx = _drefs.index(dref)
    message = b'RREF\x00' 
    message += struct.pack('<i', freq) 
    message += struct.pack('<i', idx)
    message += dref.encode() + b'\x00' * (400 - len(dref.encode()))
    res = sock.sendto(message, HOST)

def decode_message(msg):
    global _drefs

    if not msg.startswith(b'RREF,'):
        raise Exception(
                f'Message should start with "RREF,", it starts with {msg[:5]}'
                )
    res = list()
    msg = msg[5:]
    for m in batched(msg, n=8):
        m = bytes(m)
        idx, val = struct.unpack_from('<if', m)
        res.append((_drefs[idx], val))
    return res

def unsubscribe_from_dref(sock, dref):
    subscribe_to_dref(sock, dref, freq=0)

def unsubscribe_all(sock):
    global _drefs
    for d in _drefs:
        unsubscribe_from_dref(sock, d)

def get_dataref_once(sock, dref):
    subscribe_to_dref(sock, dref, freq=5)
    for _ in range(1):
        val = next(listen_to_port(sock, once=True))
        val = decode_message(val)
    unsubscribe_from_dref(sock, dref)
    return val

def get_dataref_blocking(sock, dref, freq=5):
    subscribe_to_dref(sock, dref, freq)
    for resp in listen_to_port(sock, once=False):
        val = decode_message(resp)

def listen_to_port(sock, once=False):
    # BUFFER_SIZE = 4096
    BUFFER_SIZE = 65534 # Use if more size needed. No big memory penalty.
    if once:
        data, addr = sock.recvfrom(BUFFER_SIZE)
        yield data
    else:
        while True:
            data, addr = sock.recvfrom(BUFFER_SIZE)
            yield data

if __name__ == '__main__':
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.bind(('127.0.0.1', 49001))
    if False: # Check send_command
        send_command(sock, 'sim/flight_controls/flaps_down')
        time.sleep(1)
        send_command(sock, 'sim/flight_controls/flaps_up')
        time.sleep(1)
    if False: # Check set_dataref
        set_dataref(sock, 'sim/cockpit/autopilot/heading_mag', 180, 'f')
        time.sleep(1)
        set_dataref(sock, 'sim/cockpit/autopilot/heading_mag', 0, 'f')
        time.sleep(0)
    if True: # Check for subscriptions
        print('RECEIVED:', get_dataref_once(sock, 'sim/cockpit/radios/transponder_code'))
        print('RECEIVED:', get_dataref_once(sock, 'sim/cockpit/autopilot/heading_mag'))
    unsubscribe_all(sock)



import socket
import struct
import threading
import time
from itertools import batched


class XPConnector:
    '''
    Class, that serves as the main connection between the python script and X-Plane.
    '''

    def __init__(self, host_ip=None, send_port=None, receive_port=None, listen_freq=5):
        '''
        host_ip: IP address of the machine on which X-Plane is running
        send_port: port, to which the messages are sent (on which X-Plane is listening)
            Default is 49000
        receive_port: port, which receives messages from X-Plane. Default is 49001
        listen_freq: frequency, at which a background thread checks for new messages 
        '''
        self.host_ip = host_ip if host_ip is not None else 'localhost'
        self.send_port = send_port if send_port is not None else 49000
        self.receive_port = receive_port if receive_port is not None else 49001

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.bind((self.host_ip, self.receive_port))
        self.sock.setblocking(False)

        # PRIVATE VARIABLES
        self._drefs = list()
        self._datarefs = dict()

        self._drefs_lock = threading.Lock()

        self._datarefs_lock = threading.Lock()
        self._dataref_update_condition = threading.Condition(self._datarefs_lock)
        self._stop_event = threading.Event()

        self._listener = threading.Thread(
                target=self._port_listener,
                args=(self.sock, listen_freq),
                kwargs={}
            )
        self._listener.start()


    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _get_drefs(self):
        with self._drefs_lock:
            return self._drefs.copy()


    def close(self):
        '''
        When not created with the `with` syntax, close() must be called when exiting
        the program. Otherwise, some threads will keep running indefinitely.
        '''
        for d in self._get_drefs():
            self.subscribe_to_dataref(d, freq=0)
        self._stop_event.set()
        self._listener.join()

    # PRIVATE METHODS 

    def _port_listener(self, sock, freq):
        BUFFER_SIZE = 65534
        has_updated = False
        while not self._stop_event.is_set():
            try:
                data, addr = sock.recvfrom(BUFFER_SIZE)
                with self._datarefs_lock:
                    for k, v in self._decode_message(data):
                        self._datarefs[k] = v
                has_updated = True
            except BlockingIOError:
                if has_updated: # end of data stream
                    with self._datarefs_lock:
                        self._dataref_update_condition.notify_all()
                has_updated = False
                time.sleep(1/freq)

    def _decode_message(self, msg):
        if not msg.startswith(b'RREF,'):
            raise Exception(
                    f'Message should start with "RREF,", it starts with {msg[:5]}'
                    )
        res = list()
        msg = msg[5:]
        drefs = self._get_drefs()
        for m in batched(msg, n=8):
            m = bytes(m)
            idx, val = struct.unpack_from('<if', m)
            if idx > len(drefs):
                raise Exception(
                        f'Subscriptions running in the background. pyXPUDP works'
                        ' only with no subscriptions running in the background. '
                        'Restart the simulator and ensure that no other process '
                        'is requesting datarefs from X-Plane.'
                        )
            res.append((drefs[idx], val))
        return res


    # X-PLANE API
    def send_command(self, command: str):
        '''
        Sends a command to X-Plane. 
        command: the command to send
        '''
        message = b'CMND\x00' + command.encode()
        res = self.sock.sendto(message, (self.host_ip, self.send_port))
        return res

    def set_dataref(self, dataref, value, var_type='f'):
        '''
        Sets an X-Plane dataref.
        dataref: dataref to set 
        value: value of the dataref 
        var_type: variable type of the dataref. For available types, check:
            https://docs.python.org/3/library/struct.html#format-characters
        '''
        message = b'DREF\x00' 
        message += struct.pack('=' + var_type, value) 
        message += dataref.encode() + b'\0xx'
        message += b' ' * (509 - len(message))
        res = self.sock.sendto(message, (self.host_ip, self.send_port))
        return res

    def subscribe_to_dataref(self, dataref, freq=1):
        '''
        Subscribes to a dataref. This means that X-Plane will keep sending the 
            dataref to this machine with the `freq` frequency. Note that this 
            is not the same as the `listen_freq` parameter of `XPConnector`.
            When a dataref is subscribed to, the script can use the `get_dataref`
            method to retrieve its value. To unsubscribe, send a subscription 
            request with `freq = 0`. All datarefs are automatically unsubscribed when 
            leaving the `with` block or when calling `close()`.
        dataref: dataref to subscribe to.
        freq: frequency of X-Plane sending the data
        '''
        drefs = self._get_drefs()
        if dataref not in drefs:
            with self._drefs_lock:
                drefs.append(dataref)
        idx = drefs.index(dataref)
        message = b'RREF\x00' 
        message += struct.pack('<i', freq) 
        message += struct.pack('<i', idx)
        message += dataref.encode() 
        message += b'\x00' * (400 - len(dataref.encode()))
        res = self.sock.sendto(message, (self.host_ip, self.send_port))
        return res

    def subscribe_to_datarefs(self, *datarefs, freq=1):
        '''
        Generalization of the `subscribe_to_dataref` method. Allows the script to 
            subscribe to multiple datarefs at once. 
        freq: frequency of X-Plane sending the data
        '''
        for d in datarefs:
            self.subscribe_to_dataref(d, freq)

    def get_datarefs(self, *requested_datarefs, is_blocking=True):
        '''
        Retrieves the values of the provided datarefs. If the dataref hasn't been sent 
            yet, this will block the code until it is received. This behavior can 
            be changed to returning a null value with the `is_blocking` parameter.
            If the dataref hasn't been subscribed to yet, the function automatically 
            subscribes to it. In that case, it'll be blocking regardless of the 
            `is_blocking` value.
        requested_datarefs: one or more datarefs that the user desires to retrieve.
        is_blocking: parameter that allows the user to either block the code (default)
            or to return a null value whenever the desired dataref hasn't been 
            received yet.
        Returns: list of retrieved dataref values, in order.
        '''
        drefs = self._get_drefs()
        missing_datarefs = [d for d in requested_datarefs if d not in drefs]
        if len(missing_datarefs) != 0:
            is_blocking = True 
            with self._drefs_lock:
                self._drefs.extend(missing_datarefs)
            self.subscribe_to_datarefs(*missing_datarefs)
        with self._datarefs_lock:
            while is_blocking and any(
                    d not in self._datarefs.keys() for d in requested_datarefs
                    ):
                self._dataref_update_condition.wait()
            vals = [self._datarefs.get(key, None) for key in requested_datarefs]
        return vals

    def get_dataref(self, requested_dataref, is_blocking=False):
        '''
        A simplified version of `get_datarefs` that fetches a single dataref only.
        requested_dataref: dataref the user desires to retrieve
        Returns: the value of the dataref
        '''
        return self.get_datarefs(requested_dataref, is_blocking=is_blocking)[0]
